;; Generated package description from orderless.el  -*- no-byte-compile: t -*-
(define-package "orderless" "0.8.0.20221223.161153" "Completion style for matching regexps in any order" '((emacs "26.1")) :commit "e3062280f924933e9c6f5dd1a71729ed98c8493a" :authors '(("Omar Antolín Camarena" . "omar@matem.unam.mx")) :maintainer '("Omar Antolín Camarena" . "omar@matem.unam.mx") :keywords '("extensions") :url "https://github.com/oantolin/orderless")
